# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_gfs.ipynb (unless otherwise specified).

__all__ = ['GFS']

# Cell
import os
import re
import requests
from pathlib import Path
import pandas as pd
from netCDF4 import Dataset
import numpy as np
import matplotlib.pyplot as plt

# Cell
class GFS():
    def __init__(self, path:Path, bbox:list=None, bands_sf:list=None, bands_pl:list=None,
                 last_forecast='f120', **kwargs):
        """
        path - path to save the data;
        bbox - region bounds in format [left, bottom, right top];
        bands_sf - variables for surface/mean sea level/2m/10m
        bands_pl - variables for pressure levels 1000mb up to 1mb
        last_forecast - download hourly data up to last_forecast step defined
        find_last
        """
        self.path = path
        if isinstance(self.path, str): self.path = Path(self.path)
        self.path.mkdir(exist_ok=True)
        if bbox is None: bbox = [-180, -90, 180, 90]
        self.bbox = bbox # left, bottom, right, top
        self.bands_sf = bands_sf
        self.bands_pl = bands_pl
        self.last_forecast = last_forecast
        self.find_last = True

        if not (self.path/'log.csv').is_file():
            with open(self.path/'log.csv','a') as fd:
                fd.write('date,run\n')


    def search_times(self):
        if self.find_last:
            url = f"https://nomads.ncep.noaa.gov/cgi-bin/filter_gfs_0p25_1hr.pl"
            dates_available = sorted(re.findall('dir=%2Fgfs.(.*?)">', requests.get(url).text))
            url += '?dir=%2Fgfs.' + dates_available[-1]
            runs_available = sorted(re.findall('">(.*?)</a>', requests.get(url).text))
            url += f'%2F{runs_available[-1]}'
            forecast_times = re.findall('pgrb2.0p25.(.*?)">', requests.get(url).text)
            if self.last_forecast in forecast_times:
                return dates_available[-1], runs_available[-1]
            elif len(runs_available)>1:
                return dates_available[-1], runs_available[-2]
            else:
                return dates_available[-2], '18'
        else: return dates_available, runs_available, forecast_times

    def search_files(self, tstr, run_time, forecast_hour):
        path = self.path
        left, bottom, right, top = self.bbox
        file_sf, file_pl = None, None
        if self.bands_sf is not None:
            bands_sf = ''.join([f'&var_{v}=on' for v in self.bands_sf])
            file_sf = f'https://nomads.ncep.noaa.gov/cgi-bin/' \
                   f'filter_gfs_0p25_1hr.pl?file=gfs.t{run_time}z.pgrb2.0p25.{forecast_hour}' \
                   f'&lev_10_m_above_ground=on&lev_2_m_above_ground=on&lev_mean_sea_level=on&lev_surface=on' \
                   f'{bands_sf}' \
                   f'&subregion=&leftlon={left}&rightlon={right}&toplat={top}&bottomlat={bottom}&' \
                   f'dir=%2Fgfs.{tstr}%2F{run_time}'
        if self.bands_pl is not None:
            bands_pl = ''.join([f'&var_{v}=on' for v in self.bands_pl])
            file_pl = f'https://nomads.ncep.noaa.gov/cgi-bin/' \
                   f'filter_gfs_0p25_1hr.pl?file=gfs.t{run_time}z.pgrb2.0p25.{forecast_hour}' \
                   f'&lev_1000_mb=on&lev_100_mb=on&lev_10_mb=on&lev_150_mb=on&lev_15_mb=on&lev_1_mb=on&lev_200_mb=on&lev_20_mb=on&lev_250_mb=on&lev_2_mb=on&lev_300_mb=on&lev_30_mb=on&lev_350_mb=on&lev_3_mb=on&lev_400_mb=on&lev_40_mb=on&lev_450_mb=on&lev_500_mb=on&lev_50_mb=on&lev_550_mb=on&lev_5_mb=on&lev_600_mb=on&lev_650_mb=on&lev_700_mb=on&lev_70_mb=on&lev_750_mb=on&lev_7_mb=on&lev_800_mb=on&lev_850_mb=on&lev_900_mb=on&lev_925_mb=on&lev_950_mb=on&lev_975_mb=on' \
                   f'&var_DPT=on&var_TMP=on&var_UGRD=on&var_VGRD=on' \
                   f'&subregion=&leftlon={left}&rightlon={right}&toplat={top}&bottomlat={bottom}&' \
                   f'dir=%2Fgfs.{tstr}%2F{run_time}'
        return file_sf, file_pl

    def update_log(self, tstr, run_time):
        with open(self.path/'log.csv', 'a') as f:
            f.write(f'{tstr},{run_time}\n')

    @property
    def last_log(self):
        df = pd.read_csv(self.path/'log.csv')
        if len(df)>=1:
            return df.tail(1).values.reshape(-1)
        else: return None, None


    def run(self, to_netcdf=True, delete_old=False, replace=False):
        path = str(self.path)
        tstr, run_time = self.search_times()
        log_tstr, log_run_time = self.last_log

        if not replace:
            if log_tstr is not None:
                if (str(log_tstr) == tstr) and (run_time == str(log_run_time)):
                    print('No new run is available.')
                    return False
            self.update_log(tstr, run_time)

        end_forecast = int(self.last_forecast[1:])+1
        forecast_hours = [f'f{i:03d}' for i in range(1,end_forecast)]
        if delete_old: os.system(f'cd {path}; rm *.nc')

        for f in forecast_hours:
            print(f'Downloading data for {tstr} {run_time}z {f}.')
            files = self.search_files(tstr, run_time, f)
            files = [f for f in files if f is not None]
            if len(files)>0:
                for i, file in enumerate(files):
                    if len(files)>1:
                        assert os.system(f'wget "{file}" -O {path}/GFS{run_time}z_{i}_{f}') == 0
                    else:
                        assert os.system(f'wget "{file}" -O {path}/GFS{run_time}z_{f}') == 0
                    if to_netcdf:
                        if len(files)>1:
                            assert os.system(f'cd {path}; cdo -f nc copy GFS{run_time}z_{i}_{f} GFS{run_time}z_{i}_{f}.nc') == 0
                        else:
                            assert os.system(f'cd {path}; cdo -f nc copy GFS{run_time}z_{f} GFS{run_time}z_{f}.nc') == 0
                if to_netcdf:
                    if len(files)>1:
                        os.system(f'cd {path}; cdo -O merge GFS{run_time}z_0_{f}.nc GFS{run_time}z_1_{f}.nc GFS{run_time}z_{f}.nc')
                        os.system(f'cd {path}; rm GFS{run_time}z_0_{f} GFS{run_time}z_1_{f} GFS{run_time}z_0_{f}.nc GFS_1_{f}.nc')
                    else:
                        os.system(f'cd {path}; rm GFS{run_time}z_{f}')
        return True

    def __repr__(self):
        s = ''
        for k in self.__dict__:
            s += f'{k}: {self.__dict__[k]}, '
        return s + '\n'


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_download.ipynb (unless otherwise specified).

__all__ = ['Ladsweb', 'read_log', 'update_log', 'order_status', 'download_files', 'release_order', 'order_manager',
           'run_all']

# Cell
from netCDF4 import Dataset
import matplotlib.pyplot as plt
from pathlib import Path
import pandas as pd
from rasterio.coords import BoundingBox, disjoint_bounds
import numpy as np
import json
import requests
import warnings
import re
import os
from fastprogress.fastprogress import progress_bar
from nbdev.imports import test_eq
from functools import partial
from concurrent.futures import ThreadPoolExecutor
from time import sleep
from datetime import datetime
import pdb

from .external import geturl

# Cell
class Ladsweb():
    def __init__(self, product:str, collection:str, tstart:str, tend:str,
                 bbox:list, bands:list=None, coordsOrTiles:str="coords", daynight:str="DNB",
                 repName:str='GEO', repPixSize:float=0.01, repResample:str='bilinear',
                 doMosaic:str='False', **kwargs):
        self.product, self.collection = product, collection
        self.tstart, self.tend, self.bbox, self.bands = tstart, tend, bbox, bands
        self.coordsOrTiles, self.daynight, self.repName = coordsOrTiles, daynight, repName
        self.repPixSize, self.repResample, self.doMosaic = repPixSize, repResample, doMosaic
        self._maxOrderSize, self._authFile = 1800, os.path.expanduser('~/.ladsweb')

    @property
    def _email(self):
        with open(self._authFile, 'r') as f:
            data = json.load(f)['email']
        assert len(data) > 0
        assert '@' in data
        return data

    @property
    def _key(self):
        with open(self._authFile, 'r') as f:
            data = json.load(f)['key']
        assert len(data) > 0
        return data

    def search_files(self):
        "Search for files for the product, region and time span given."
        url = (f"https://modwebsrv.modaps.eosdis.nasa.gov/axis2/services/MODAPSservices/" +
            f"searchForFiles?product={self.product}&collection={self.collection}&" +
            f"start={self.tstart}&stop={self.tend}&north={self.bbox[3]}&south={self.bbox[1]}" +
            f"&west={self.bbox[0]}&east={self.bbox[2]}&coordsOrTiles={self.coordsOrTiles}" +
            f"&dayNightBoth={self.daynight}")
        return re.findall('<return>(.*?)</return>', requests.get(url).text)

    def order_size(self):
        "Calculates the number of files in the order."
        if self.bands is None:
            raise Exception("`bands` list required to calculate order_size.")
        return len(self.search_files())*len(self.bands)

    def split_times(self, maxOrderSize=None):
        "Split a single order into multiple orders if the order size is too large."
        if maxOrderSize is None: maxOrderSize = self._maxOrderSize
        order_size = self.order_size()
        if order_size <= maxOrderSize:
            return [self]
        n_splits = order_size // maxOrderSize + 1
        times = pd.date_range(self.tstart, self.tend)
        bk = -(len(times) % n_splits)
        if bk == 0: bk = None
        splits = np.split(times[:bk], n_splits)
        splits[-1] = splits[-1].append(times[bk:])
        times = [(str(t[0]), str(t[-1]+pd.Timedelta(days=1-1e-5).round('s'))) for t in splits]
        tstart, tend = zip(*times)
        kwargs = self.__dict__
        group = []
        for ti, tf in zip(tstart, tend):
            kwargs['tstart'], kwargs['tend'] = ti, tf
            group.append(Ladsweb(**kwargs))
        return group

    def send_order(self, ids):
        "Send order for a set of ids obtained with `search_files` method."
        ids = ','.join(ids)
        bands = ','.join([self.product + f'___{b}' for b in self.bands])
        url = (f"http://modwebsrv.modaps.eosdis.nasa.gov/axis2/services/MODAPSservices/" +
            f"orderFiles?fileIds={ids}" +
            f"&subsetDataLayer={bands}" +
            f"&geoSubsetNorth={self.bbox[3]}" +
            f"&geoSubsetSouth={self.bbox[1]}" +
            f"&geoSubsetEast={self.bbox[2]}" +
            f"&geoSubsetWest={self.bbox[0]}" +
            f"&reprojectionName={self.repName}" +
            f"&reprojectionOutputPixelSize={self.repPixSize}" +
            f"&reprojectionResampleType={self.repResample}" +
            f"&doMosaic={self.doMosaic}" +
            f"&email={self._email}")
        return re.findall('<return>(.*?)</return>', requests.get(url).text)[0]

    def run(self, path_save):
        "Send request and update log file."
        ids = self.search_files()
        if len(ids) == 0:
            warnings.warn("No files found", UserWarning)
            return
        if self.bands is None: raise Exception("A list of `bands` is not defined")
        orderId = self.send_order(ids)
        status = order_status(orderId)
        update_log(Path(path_save)/'order_log.json', orderId, status)
        print(f'New request sent with orderId {orderId}')

    def __repr__(self):
        s = ''
        for k in self.__dict__:
            s += f'{k}: {self.__dict__[k]}, '
        return s + '\n'

# Cell
def read_log(file):
    "Read log file."
    with open(file, 'r') as f:
        data = json.load(f)
    return data

def update_log(file, orderId, status):
    "Update log file."
    now = datetime.now()
    current_time = now.strftime("%Y-%m-%d_%H:%M:%S")
    if file.is_file():
        stored_data = read_log(file)
        if orderId not in [k for k in stored_data]:
            stored_data[orderId] = {'status': '', 'time': ''}
        if stored_data[orderId]['status'] != status:
            stored_data[orderId]['status'] = status
            stored_data[orderId]['time'] = current_time
    else:
        stored_data = {orderId: {'status': status, 'time': current_time}}

    with open(file, 'w') as f:
        json.dump(stored_data, f)

def order_status(orderId):
    "Check order status."
    url = (f"http://modwebsrv.modaps.eosdis.nasa.gov/axis2/services/MODAPSservices/" +
            f"getOrderStatus?orderId={orderId}")
    return re.findall('<return>(.*?)</return>', requests.get(url).text)[0]

def download_files(orderId, path_save, auth=None):
    "Download files if the order is Available."
    if auth is None: raise Exception("`auth` code is not defined")
    status = order_status(orderId)
    if status != 'Available':
        msg = f"Order is not Available, current status is {status}"
        warnings.warn(msg, UserWarning)
        return
    url = f'https://ladsweb.modaps.eosdis.nasa.gov/archive/orders/{orderId}'
    files = pd.DataFrame(json.loads(geturl(url + '.json', auth)))
    checksums = geturl(url + f'/checksums_{orderId}', auth)
    hdfs = re.findall('(.*?.hdf)', checksums)
    ch = [tuple([k for k in h.split(' ') if k != '']) for h in hdfs]
    check_df = pd.DataFrame(ch, columns=['checksum', 'size', 'name'])
    check_df = check_df.drop('size', axis=1)
    files = pd.merge(files, check_df, how='left', on='name')
    files['verified'] = False

    for i in progress_bar(range(len(files))):
        file, checksum = files.loc[i, ['name', 'checksum']]
        csum = None
        if (Path(path_save)/file).is_file():
            csum = os.popen(f'cksum {str(path_save)}/{file}').read().split(' ')[0]
        if csum is None or checksum != csum:
            n_tries = 0
            while ~files.loc[i, 'verified'] and n_tries<5:
                #print(f'Downloading {file}')
                with open(Path(path_save)/f'{file}', mode='w+b') as fh:
                    try: geturl(f'{url}/{file}', auth, fh)
                    except: warnings.warn(f'Unable to get {url}/{file}', UserWarning)
                csum = os.popen(f'cksum {str(path_save)}/{file}').read().split(' ')[0]
                if str(checksum) == 'nan': checksum = csum
                files.loc[i, 'verified'] = checksum == csum
                n_tries += 1
        elif checksum == csum: files.loc[i, 'verified'] = True
    log_file = f'download_log_{orderId}.csv'
    files.to_csv(Path(path_save)/log_file)
    not_verified = np.sum(~files.verified)
    if not_verified > 0:
        msg = f"Checksum failed for {not_verified} files. Check the {log_file}."
        warnings.warn(msg, UserWarning)
    return not_verified

def release_order(orderId, email=None):
    "To release order after download the files."
    if email is None: raise Exception("`email` is not defined")
    url = (f"http://modwebsrv.modaps.eosdis.nasa.gov/axis2/services/MODAPSservices/" +
    f"releaseOrder?orderId={orderId}&email={email}")
    status = re.findall('<return>(.*?)</return>', requests.get(url).text)[0]
    return status == '1'

def order_manager(path_save):
    "Manage active orders in log file. Check the status and download the files for each order"
    authFile = os.path.expanduser('~/.ladsweb')
    with open(authFile, 'r') as f:
        f = json.load(f)
        email = f['email']
        auth = f['key']

    log_file = Path(path_save)/'order_log.json'
    while True:
        data = read_log(log_file)

        # Update status
        for orderId in data:
            if data[orderId]['status'] != 'Complete':
                status = order_status(orderId)
                update_log(log_file, orderId, status)

        # Download if available (wait 10 min)
        now = datetime.now()
        current_time = now.strftime("%Y%m%d_%H%M%S")
        for orderId in data:
            if data[orderId]['status'] == 'Available':
                d = (pd.Timestamp(' '.join(current_time.split('_')))
                     - pd.Timestamp(' '.join(data[orderId]['time'].split('_')))).seconds
                if d//60 > 10:
                    status = download_files(orderId, path_save, auth)
                    if status == 0:
                        result = release_order(orderId, email)
                        status = 'Complete' if result else 'One or more files not verified'
                        update_log(log_file, orderId, status)
                        print(f'Files for order {orderId} saved at {path_save}.')

        # Check if stop
        n = 0
        for orderId in data:
            if data[orderId]['status'] in ['Complete', 'One or more files not verified',
                                           'Canceled', 'Removed']:
                n += 1
        stop = len(data) == n
        if stop: return
        sleep(20)

def run_all(request_list, path_save):
    "Send a list of requests and initiate order manager."
    for request in request_list:
        request.run(path_save)
        sleep(5)
    order_manager(path_save)